/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef WRAP_FFTW_HH
#define WRAP_FFTW_HH

#include "wrapfft.hh"
#include "fftw3.h"

class fftw_plan_pointer : public plan_pointer {
public:
    fftw_plan_pointer(void* plan=0, 
		      wrapfft::fft_dirtype typ=wrapfft::fft_float_fcomplex);
    ~fftw_plan_pointer(void);
    fftw_plan_pointer* clone(void);
    void kill(void);
    void* release(void);
    void reset(void*);
    fftwf_plan get_fftwf_plan(void);
    fftw_plan  get_fftwd_plan(void);
private:
    wrapfft::fft_dirtype _type;
    void* _plan;
};

/**  The wrapfftw class implements the standard interface for the fftw3.
  *  package.
  *  \brief fftw3 package wrapper.
  *  \author J. Zweizig
  *  \version 1.0: Last modified April 10, 2008
  */
class wrapfftw : public wrapfft {
public:
    /**  The default constructor is all that's needed.
      *  \brief Default constructor.
      */
    wrapfftw(void);

    /**  Virtual destructor to shut down packages.
      *  \brief Default destructor.
      */
    ~wrapfftw(void);

    /**  Optional entry to fetch saved plans from a specified file. This
      *  method defaults to a null routine, and must be overridden if 
      *  an appropriate funtion is available.
      *  \brief Fetch precalculated plans
      *  \param file  File name.
      */
    void fetch_plans(const std::string& file);

    /**  Select an fft length that will lead to a fast transform.
      *  \brief Chose an fft length that will be fast
      *  \param N Starting length.
      *  \return recommended length.
      */
    long pick_length(long N) const;

    /**  Transform N real data points to a single-sided complex dft.
      *  The input vector must contain N float elements and the output
      *  vector must accomodate N+1 complex elements.
      *  \brief DFT of real data.
      *  \param N   Length of input series.
      *  \param in  Pointer to real (float) series.
      *  \param out Pointer to storage for N+1 element complex series.
      */
    void transform(long N, const float* in, std::complex<float>* out);

    /**  Perform an inverse DFT from a single-sided complex series to
      *  a real series.
      *  The input vector contains N/2+1 complex elements and the output
      *  vector must accomodate N float elements.
      *  \brief Inverse DFT of single-sided data.
      *  \param N   Length of real (output) series.
      *  \param in  Pointer to an N/2+1 element complex series.
      *  \param out Pointer to storage for an N element (float) series.
      */
    void transform(long N, const std::complex<float>* in, float* out);

    /**  Perform a forward or backward DFT of N complex data points. The
      *  resulting series overwrites the input data. The length is set 
      *  -N to indicate an backward transform.
      *  The input/output vector must contain N complex elements.
      *  \brief DFT of real data.
      *  \param N   Length of input series.
      *  \param in  Pointer to real (float) series.
      *  \param out Pointer to storage for N element complex series.
      */
    void transform(long N, std::complex<float>* out);

    /**  Transform N real data points to a single-sided complex dft.
      *  The input vector must contain N double elements and the output
      *  vector must accomodate N/2+1 complex elements.
      *  \brief DFT of real data.
      *  \param N   Length of input series.
      *  \param in  Pointer to real (float) series.
      *  \param out Pointer to storage for N/2+1 element complex series.
      */
    void transform(long N, const double* in, std::complex<double>* out);

    /**  Perform an inverse DFT from a single-sided complex series to
      *  a real series.
      *  The input vector contains N/2+1 complex elements and the output
      *  vector must accomodate N float elements.
      *  \brief Inverse DFT of single-sided data.
      *  \param N   Length of real (output) series.
      *  \param in  Pointer to an N/2+1 element complex series.
      *  \param out Pointer to storage for an N element (float) series.
      */
    void transform(long N, const std::complex<double>* in, double* out);

    /**  Perform a forward or backward DFT of N complex data points. The
      *  resulting series overwrites the input data. The length is set 
      *  -N to indicate an backward transform.
      *  The input/output vector must contain N complex elements.
      *  \brief DFT of real data.
      *  \param N   Length of input series.
      *  \param out Pointer to storage for N element complex series.
      */
    void transform(long N, std::complex<double>* out);

protected:
    /**  Optional entry to fetch saved plans from a specified file. This
      *  method defaults to a null routine, and must be overridden if 
      *  an appropriate funtion is available.
      *  \brief Fetch precalculated plans
      *  \param type  Direction, fft type and data type.
      *  \param N     Number of elements in series.
      */
    fftw_plan_pointer* find_plan(plan_id id);

private:
    thread::readwritelock mux;
};

inline fftwf_plan 
fftw_plan_pointer::get_fftwf_plan(void) {
    return reinterpret_cast<fftwf_plan>(_plan);
}

inline fftw_plan
fftw_plan_pointer::get_fftwd_plan(void) {
    return reinterpret_cast<fftw_plan>(_plan);
}

#endif // WRAP_FFT_HH
