/*
 *    DMTifc.h  provides a C callable interface for those wishing to
 *    use C code in the data monitor. THe hooks will have to be added
 *    to make any functionality availeble in the C++ class.
 */
#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#include "Csum.h"
#include "DMTifc.h"

#define DSSMAXEVT 10
#define DSSMAXMSG 30
#define MAXADC 2048

char *DSSversion = "v1.2 Nov 6, 00";

struct DSSparam{
  int framePerFile;       /* number of frame per file                   */
  int freqLSC;            /* frequency for the LSC channel in kHz       */
  double scaleLSC;        /* scaling factor for the LSC channel         */
  int htmlPeriod;         /* update period for the html status report   */
  Csum *cs;               /* channel summary object                     */
  int nFrame;             /* number of procesed frames                  */
  FrFile *file;           /* pointer to the frameFile                   */
  char frFName[256];      /* frame file name                            */
  FILE *log;              /* pointer to the log file                    */
  char *logDir;           /* directory for the report and log files     */ 
  char *dataDir;          /* directory for the data files               */ 
  int nEvtMsg;            /* number of event messages                   */
  int nLogMsg;            /* number of log messages                     */
  char *evtMsg[DSSMAXEVT];/* hold event messages                        */
  char *logMsg[DSSMAXMSG];/* hold log messages                          */
  int  ssRate;            /* spectrum saving rate                       */
  char config[2048];      /* hold configuration used                    */
  double badRate[MAXADC]; /* trigger rate for a channel (size nAdc)     */
  double chi2[MAXADC];    /* chi2 for a channel(size nAdc)(see DSSevent)*/
  double threshold;       /* threshol to considere a channel as strange */
  int nchevt;             /* number of channel above threshold for an evt*/
  FRBOOL verbose;         /* if no supress the printf                   */
};

struct DSSparam DSS;

/**--------------------------------------------------------------------------*/
void DMT_Initialize(int argc, const char *argv[]) {
/**--------------------------------------------------------------------------*/
/* options are:                                                              */
/*   -n number of frames par file (default: 1000)                            */
/*   -f frequency (in kHz for the main channel)                              */
/*   -s scaling factor for the main channel for the float ->int conversion   */
/*   -ld directory for the reports                                           */
/*   -dd directory for the reports                                           */
/*   -p  update period for the html report (default 5 s)                     */
/**--------------------------------------------------------------------------*/
 int i;

           /*-------------------- Default values ----------------------------*/

 DSS.framePerFile  = 2000;
 DSS.freqLSC  = 4;
 DSS.scaleLSC = 1.;
 DSS.nFrame   = 0;
 DSS.verbose  = FR_NO;
 DSS.htmlPeriod = 10;
 DSS.nLogMsg = 0;
 DSS.file    = NULL;
 DSS.ssRate  = 1000;
 DSS.threshold = 2.5;
 DSS.nchevt    = 8;
 FrStrCpy(&(DSS.logDir),"/export/home/mours/log");
 FrStrCpy(&(DSS.dataDir),"/export/home/mours/data");

           /*------------------- Loop over the parameters -------------------*/

 for (i=1; i<argc; i=i+2)
   {if      (strcmp(argv[i],"-dd") == 0) {FrStrCpy(&(DSS.dataDir),(char*)argv[i+1]);} 
    else if (strcmp(argv[i],"-ld") == 0) {FrStrCpy(&(DSS.logDir), (char*)argv[i+1]);} 
    else if (strcmp(argv[i],"-n")  == 0) {DSS.framePerFile = atoi(argv[i+1]);}
    else if (strcmp(argv[i],"-v")  == 0) {DSS.verbose = FR_YES;}
    else if (strcmp(argv[i],"-f")  == 0) {DSS.freqLSC      = atoi(argv[i+1]);}
    else if (strcmp(argv[i],"-n")  == 0) {DSS.framePerFile = atoi(argv[i+1]);}
    else if (strcmp(argv[i],"-et") == 0) {DSS.threshold    = atof(argv[i+1]);}
    else if (strcmp(argv[i],"-ec") == 0) {DSS.nchevt       = atoi(argv[i+1]);}
    else if (strcmp(argv[i],"-p")  == 0) {DSS.htmlPeriod   = atoi(argv[i+1]);}
    else if (strcmp(argv[i],"-ssr")== 0) {DSS.ssRate       = atoi(argv[i+1]);}
    else
       {printf("error while reading parameters\n");}}

           /*------------------ Check start/stop parameters------------------*/

 sprintf(DSS.config,
     "  DSSbldr version      : %s\n"
     "  # of frame per file  : %d\n"
     "  # of frame per file  : %d\n"
     "  LSC channelfrequency : %dkHz\n"
     "  LSC channel scaling  : %f\n"
     "  Data   directory     : %s\n"
     "  Report directory     : %s\n"
     "  Spectrum saving rate : %d\n"
     "  Event definition     : %.2f %d\n",
	 DSSversion,    DSS.framePerFile, DSS.framePerFile, DSS.freqLSC,
         DSS.scaleLSC,  DSS.dataDir,      DSS.logDir,        DSS.ssRate,
         DSS.threshold, DSS.nchevt);

 printf("-----------Parameters used--------------\n%s",DSS.config);
 printf("----------------------------------------\n");
 printf(" FFTW words size is %d bytes\n",fftw_sizeof_fftw_real());
 printf("----------------------------------------\n");
 
           /*---------------- create channel summary object -----------------*/

 DSS.cs = CsumNew(32, 32000,0.,"H");

 for(i=0; i<MAXADC; i++) {DSS.badRate[i] = 1.;}

 return;
}
/**--------------------------------------------------------------------------*/
void DMT_Halt(void) {
/**--------------------------------------------------------------------------*/
    printf("Monitor has completed\n");
     
    FrFileOEnd(DSS.file);
}

/*---------------------------------------------------------------------------*/
void DSSlog(struct FrameH* frame, char* format, ...) {
/*---------------------------------------------------------------------------*/
 int i;
 va_list    args;
 char       text[1024], mytime[50];
 time_t tp;

 va_start (args, format);
 vsprintf (text, format, args);
 va_end (args);

 i = DSS.nLogMsg%DSSMAXMSG;

 if(DSS.logMsg[i] == NULL) DSS.logMsg[i] = malloc(1024);
 if(DSS.logMsg[i] == NULL) return;
 
 tp = time(NULL);
 strcpy(mytime, ctime(&tp));
 mytime[20] = '\0';
 sprintf(DSS.logMsg[i]," (%d %s) %s",frame->GTimeS, &(mytime[11]), text);

 if(DSS.verbose == FR_YES) printf("\n >>> %s\n",DSS.logMsg[i]);

 fprintf(DSS.log,"%s\n",DSS.logMsg[i]);
 fflush(DSS.log);

 DSS.nLogMsg++;

 return;}

/*---------------------------------------------------------------------------*/
void DSSevent(struct FrameH* frame) {
/*---------------------------------------------------------------------------*/
 int i, imsg, j, imax, count;
 char   *p, mytime[50];
 time_t tp;
 double chi2, w, max, decay;
 FrSummary *sumChi2;
 FrVect *names;

 if((sumChi2 = FrSummaryFind(frame,"H:CsFChi2")) == NULL) return;
 if((names = FrStatDataGetV(frame,"H:CsFNames")) == NULL) return;

 count = 0;
 decay = .99;

 for(i=0; i<sumChi2->moments->nData; i++)
   {chi2 = 0.01*sumChi2->moments->dataS[i];
    DSS.chi2[i] = 0.;
                 /*------------ update bad channel list -------------------*/

    if(chi2 > DSS.threshold) 
          w = 1.;
     else w = 0.;
     DSS.badRate[i] = decay*DSS.badRate[i] +(1.-decay)*w;

                   /*------------- count the number of trigger -----*/

     if(DSS.badRate[i] > .2) continue;
     if(chi2 < DSS.threshold) continue; 
     count ++;
     DSS.chi2[i] = chi2;} 

 if(count < DSS.nchevt) return;

                  /*------- fill message line -------*/

 imsg = DSS.nEvtMsg%DSSMAXEVT;

 if(DSS.evtMsg[imsg] == NULL) DSS.evtMsg[imsg] = malloc(1024);
 if(DSS.evtMsg[imsg] == NULL) return;
 
 tp = time(NULL);
 strcpy(mytime, ctime(&tp));
 mytime[20] = '\0';
 p = DSS.evtMsg[imsg];
 p += sprintf(p," <font color=\"#006600\">(%d %s) </font>"
                 "<font color=\"#1822CD\">%d </font>channels:", 
                frame->GTimeS, &mytime[11],count);

 for(j=0; j<count; j++)
   {if(j > 7)
      {p += sprintf(p,"...");
       break;}
    max = 0;
    imax = 0;
    for(i= 0; i<sumChi2->moments->nData; i++)
      {if(DSS.chi2[i] < max) continue;
       max = DSS.chi2[i];
       imax = i;}
    if(names->dataQ[imax] != NULL)
        p += sprintf(p," %s<font color=\"#1822CD\">(%.2f, %.2f)</font>",
                   names->dataQ[imax],DSS.chi2[imax], DSS.badRate[imax]);
    else{
        printf("\nWarning dataQ= NULL for imax=%d (ndata=%d)\n",
          imax,sumChi2->moments->nData);
      }
   DSS.chi2[imax] = 0.;}

 DSS.nEvtMsg++;

 return;}

/*-------------------------------------------------------------- CsumStatus--*/
void DSSstatus(FrameH *frame, double mean)
/*---------------------------------------------------------------------------*/
{FILE *fp;
 char *filehtml="/opt/CDS/a/apacheFiles/htdocs/gds/monitor_reports/DSSbldr/DSSbldr_status.html";
 time_t tp;
 int index,i;
 FrSummary *sumChi2;
 FrVect *names;

 if((sumChi2 = FrSummaryFind(frame,"H:CsFChi2")) == NULL) return;
 if((names = FrStatDataGetV(frame,"H:CsFNames")) == NULL) return;

 DSSevent(frame);

 if(frame->GTimeS % DSS.htmlPeriod != 0) return;

 fp = fopen(filehtml,"w");
 if(fp == NULL)
   {printf(" DDSbldr_status.html file open error\n"); 
   return;}

 tp = time(NULL); 
 fprintf(fp,"<html>\n<head>\n"
       "<title>Data Summary Set Builder Status</title>\n</head>\n<body>\n"
       "<center><b><font color=\"#1822CD\"><font size=+3>DSS Builder Status"
       "</font></font></b></center>\n"
       "<hr WIDTH=\"100%%\">\n  <b>\n"
       "<br>Report time: <font color=\"#1822CD\">%s</font> "
       "(update every %d s.)\n"
       "<br>Last frame: <font color=\"#1822CD\">%d</font>"
       "     (mean chisquare=<font color=\"#1822CD\">%.2f</font>)\n",
        ctime(&tp), DSS.htmlPeriod, frame->GTimeS, mean*.01);

 fprintf(fp,
       "<br>Frame processed: <font color=\"#1822CD\">%d</font>\n"
       "<br>Current output file: %s (<font color=\"#1822CD\">%d</font> frames,"
       " <font color=\"#1822CD\">%.3f</font> MBytes)\n"
       "<br>Number of fast/slow channels: %d/%d"
       "   Resample H2:LSC-AS_Q channel at %d kHz (scale=%.2f)"
       "   Frames per file:%d\n"
       "<hr WIDTH=\"100%%\">\n"
       "<b><font color=\"#1822CD\"><font size=+1>Recent Events " 
       "(out of %d events):</font></font></b>\n"
       "<ul>\n<font size=-1>",
         DSS.nFrame, DSS.frFName, 
         DSS.file->nFrames,  DSS.file->nBytes/1048576.,
         DSS.cs->chi2->nData, DSS.cs->slowRes->sum->nData, DSS.freqLSC, DSS.scaleLSC,
         DSS.framePerFile,DSS.nEvtMsg);

  /*------- Printf event list -----*/

 for(i=DSS.nEvtMsg-1; i>=0; i--)
     {if(i<DSS.nEvtMsg-DSSMAXEVT) break;
      index = i%DSSMAXEVT;
      fprintf(fp,"<li><b>%s</b></li>\n",DSS.evtMsg[index]);}

 fprintf(fp,"</font></ul>\n"
       "<hr WIDTH=\"100%%\">\n"
       "<b><font color=\"#1822CD\"><font size=+1>List of loud channels: " 
	 "</font></font></b><br><font size=-1>\n");

       /*----- printf bad channel list -----------------*/

 for(i=0; i<sumChi2->moments->nData; i++)
   {if(DSS.badRate[i] < .3) continue;
   if(names->dataQ[i] == NULL) continue;
    fprintf(fp," %s<font color=\"#1822CD\">(%.2f)</font>",
	   names->dataQ[i],DSS.badRate[i]);}

 fprintf(fp,"</ul>\n"
       "</font>\n<hr WIDTH=\"100%%\">\n"
       "<b><font color=\"#1822CD\"><font size=+1>Recent Log messages " 
       "(out of %d log messages):</font></font></b>\n"
       "<ul>\n<font size=-1>",DSS.nLogMsg);

  /*------- Printf Log Messages -----*/

 for(i=DSS.nLogMsg-1; i>=0; i--)
     {if(i<DSS.nLogMsg-DSSMAXMSG) break;
      index = i%DSSMAXMSG;
      fprintf(fp,"<li><b>%s</b></li>\n",DSS.logMsg[index]);}

 fprintf(fp,"</ul>\n");
 
   /*- end of file ----*/

 fprintf(fp,"</ul>\n<hr WIDTH=\"100%%\"><div align=right>"
	 "For any question, e-mail to mours@ligo.caltech.edu"
         "</div>\n</body>\n</html>");

 fclose(fp);

 return;}
 
/**--------------------------------------------------------------------------*/
void DMT_Frame(const void* f) {
/**--------------------------------------------------------------------------*/
struct FrameH* frame = (struct FrameH*)f;
FrAdcData *adc;
FrVect *vect;
struct FrProcData *proc;
struct FrSummary *sumChi2;
static unsigned int lastGPS=0;
static int nDecimate;
double mean;
char lName[512], program[256];
int nFrames, nBytes;

         /*------------------- file close/open ------------------------------*/

 if((DSS.file == NULL) || (frame->GTimeS % DSS.framePerFile == 0))
    {if(DSS.file != NULL) 
        {nFrames = DSS.file->nFrames;
         nBytes = DSS.file->nBytes;
         FrFileOEnd(DSS.file);
	 fclose(DSS.log);}
     else
        {nFrames = 0;
         nBytes = 0;}

     sprintf(lName,"%s/H-%d.log",DSS.logDir,frame->GTimeS);
     DSS.log = fopen(lName, "w");
     if(DSS.log == NULL)
          {DSSlog(frame," Error: log file open error");
           exit(0);}

     sprintf(DSS.frFName,"%s/H-%d.DSS",DSS.dataDir,frame->GTimeS);
     DSSlog(frame," Open frame file:%s "
            "(previous file: %d frames %.2f MBytes)",
            DSS.frFName, nFrames, nBytes/1048576.);
     sprintf(program,"DSSbldr %s",DSSversion);
     DSS.file = FrFileONewH(DSS.frFName, 6, program);
     if(DSS.file == NULL)
          {DSSlog(frame," Error: frame file open error");
           exit(0);}}

        /*------------------ check if we miss some frame --------------------*/

 DSS.nFrame ++;
 if((lastGPS+1 != frame->GTimeS) && (lastGPS > 0))
   {DSSlog(frame,">>> Warning: %d frame(s) missing<<<",
                      frame->GTimeS - lastGPS -1);}
 lastGPS = frame->GTimeS;

         /*---------------- do frame processing -----------------------------*/

 CsumProc(DSS.cs, frame);

 sumChi2 = FrSummaryFind(frame,"H:CsFChi2");
 FrvMean(sumChi2->moments, &mean);
 if(DSS.verbose == FR_YES) 
     {printf("GTimeS=%d frame processed:%6d <chi2>=%6.2f\r",
        frame->GTimeS, DSS.nFrame, mean*.01);
      fflush(stdout);}
 if(frame->GTimeS % 100 == 0) {fprintf(DSS.log,
        "GTimeS=%d frame processed:%6d <chi2>=%6.2f\r",
          frame->GTimeS, DSS.nFrame, mean*.01);
         fflush(DSS.log);}


        /*-------- build proc data: decimate copy of the asymetric port----*/

 if(DSS.freqLSC != 0)
   {adc  = FrAdcDataFind(frame,"H2:LSC-AS_Q");
    if(adc != NULL)
      {if(DSS.scaleLSC != 1.) FrvScale(DSS.scaleLSC, adc->data, adc->data, NULL);
       nDecimate = adc->data->nData/(1024*DSS.freqLSC);
       if(DSS.freqLSC < 0)
          {proc = FrProcDataNew(frame,adc->name, 
                         adc->sampleRate, adc->data->nData, adc->nBits);
           memcpy(proc->data->data, adc->data->data, adc->data->nBytes);
           FrvDecimate(proc->data, nDecimate);
	   vect = FrvCopyToI(proc->data, adc->name);
	   FrVectFree(proc->data);
           proc->data = vect;
           proc->sampleRate = proc->sampleRate/nDecimate;
           FrameTagAdc(frame,"NONE");}
        else
          {FrvDecimate(adc->data, nDecimate);
	   vect = FrvCopyToI(adc->data, adc->name);
	   FrVectFree(adc->data);
           adc->data = vect;
           adc->sampleRate = adc->sampleRate/nDecimate;
           FrameTagAdc(frame,"H2:LSC-AS_Q");}}}


        /*----------- save spectrum if requested---------------------------*/

 if((DSS.nFrame == 1) || 
    (frame->GTimeS % DSS.ssRate == 0)) CsumSaveS(DSS.cs, frame);
  
        /*----------- write data (no Raw data)---------------------------*/

 FrameWrite(frame, DSS.file);
 
 DSSstatus(frame,mean);

 return;}

